# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

# Version of the lazygrid package
__version__ = "4.0.0"

from . import logger, datasets, database, statistics, neural_models, \
    grid, model_selection, plotter, wrapper

logger.initialize_logging()



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import logging
import os
from logging.handlers import RotatingFileHandler


log_info = logging.getLogger("lazygrid-debug")
log_info.addHandler(logging.NullHandler())
log_warn = logging.getLogger("lazygrid-warning")
log_warn.addHandler(logging.NullHandler())


def initialize_logging(path: str = "./log", log_name: str = "default-logger",
                       date: bool = True, output_console: bool = False) -> logging.Logger:
    """
    Initialize log file handler.

    Parameters
    ----------
    :param path: location where the log file will be saved
    :param log_name: log file name
    :param date: if True the current datetime will be put before the log file name
    :param output_console: if True it prints log messages to the default console
    :return: log file handler
    """

    if date:
        log_name = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S") + "-" + log_name
    log_name = os.path.join(path, log_name)

    # create log folder if it does not exists
    if not os.path.isdir(path):
        os.mkdir(path)

    # remove old logger if it exists
    if os.path.exists(log_name):
        os.remove(log_name)

    # log file format
    formatter = logging.Formatter("[%(levelname)s %(asctime)s] %(message)s", "%Y-%m-%d %H:%M:%S")

    # create logger instances
    log_info.setLevel(logging.DEBUG)
    log_warn.setLevel(logging.WARNING)

    # the 'RotatingFileHandler' object implements a log file that is automatically limited in size
    # debug log handler
    fh_info = RotatingFileHandler(log_name + "-debug.log",
                                  mode='a',
                                  maxBytes=100 * 1024 * 1024,
                                  backupCount=2,
                                  encoding=None,
                                  delay=0)
    fh_info.setLevel(logging.DEBUG)
    fh_info.setFormatter(formatter)
    log_info.addHandler(fh_info)

    # warning log handler
    fh_warning = RotatingFileHandler(log_name + "-warning.log",
                                     mode='a',
                                     maxBytes=100 * 1024 * 1024,
                                     backupCount=2,
                                     encoding=None,
                                     delay=0)
    fh_warning.setLevel(logging.WARNING)
    fh_warning.setFormatter(formatter)
    log_warn.addHandler(fh_warning)

    if output_console:
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)
        log_info.addHandler(ch)

    log_info.debug("Starting " + log_name + "!")

    return log_info


def close_logging(logger: logging.Logger):
    """
    Close log file handler.

    Parameters
    ----------
    :param logger: log file handler
    :return: None
    """

    for handler in logger.handlers:
        handler.close()
        logger.removeHandler(handler)



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import time
import traceback
import numpy as np
import pandas as pd
import os
import functools
from typing import Union, Callable, List
from logging import Logger
import pycm
from scipy.stats import mannwhitneyu
from statsmodels.stats.proportion import proportion_confint
from sklearn.model_selection import StratifiedKFold
from sklearn.datasets import make_classification
from sklearn.linear_model import RidgeClassifier, LogisticRegression
from tqdm import tqdm
from .statistics import find_best_solution, confidence_interval_mean_t
from .wrapper import Wrapper
from .plotter import generate_confusion_matrix
from .logger import log_info


def cross_validation(model: Wrapper,
                     x: np.ndarray, y: np.ndarray,
                     x_val: np.ndarray = None, y_val: np.ndarray = None,
                     random_data: bool = True, random_model: bool = True,
                     seed: int = 42, n_splits: int = 10, score_fun: Callable = None,
                     generic_score: Callable = None, bar_position: int = 0,
                     verbose: bool = True) -> (dict, list, List[np.ndarray], List[np.ndarray]):
    """
    Apply cross-validation on the given model.

    Examples
    --------
    >>> from sklearn.linear_model import LogisticRegression
    >>> from sklearn.datasets import make_classification
    >>> import lazygrid as lg
    >>>
    >>> x, y = make_classification()
    >>>
    >>> lg_model = lg.wrapper.SklearnWrapper(LogisticRegression())
    >>> score, fitted_models, y_pred_list, y_list = lg.model_selection.cross_validation(model=lg_model, x=x, y=y)
    >>>
    >>> type(score)
    <class 'dict'>
    >>> type(fitted_models)
    <class 'list'>

    Notes
    --------
    The cross-validation can be applied in three different ways, depending on the
    user needs or the application:
        - generating random training and validation sets at each iterations (random_data=True)
        - initializing randomly the given model at each iterations (random_model=True)
        - applying both the previous options

    The second option is usually recommended for deep neural networks and big data sets.

    Parameters
    --------
    :param model: machine learning model
    :param x: input data
    :param y: input labels
    :param x_val: validation data
    :param y_val: validation labels
    :param random_data: if True it enables data randomization
    :param random_model: if True it enables model randomization (if applicable)
    :param seed: seed used to make results reproducible
    :param n_splits: number of cross-validation iterations
    :param score_fun: sklearn-like scoring function used to evaluate the model performance
    :param generic_score: generic score function
    :param bar_position: progress bar position
    :param verbose: print progress bars
    :return: cross-validation scores, fitted models, list of predicted labels and true labels
    """

    # Check input parameters
    assert score_fun or generic_score or hasattr(model, "score")
    assert isinstance(x, np.ndarray) and isinstance(y, np.ndarray)
    assert isinstance(random_data, bool)
    assert isinstance(random_model, bool)
    assert isinstance(seed, int)
    assert isinstance(n_splits, int)
    if not random_data:
        assert x_val is not None and y_val is not None

    if random_model and not random_data:
        assert isinstance(x_val, np.ndarray) and isinstance(y_val, np.ndarray)

    # Useful variables
    fitted_models = []
    y_pred_list = []
    y_list = []
    score = {}

    # prepare data for cross-validation
    if random_data:
        skf = StratifiedKFold(n_splits=n_splits, shuffle=True, random_state=seed)
        list_of_splits = [split for split in skf.split(x, y)]

    # Cross validation
    log_info.info("Start cross-validation")
    splits = np.arange(0, n_splits)
    if verbose:
        splits = tqdm(np.arange(0, n_splits), desc="Cross-validation split", position=bar_position, leave=True)

    for split_index in splits:

        log_info.info("Split %d" % split_index)

        # randomize data if needed
        if random_data:
            train_index, val_index = list_of_splits[split_index]
            x_train, x_val = x[train_index], x[val_index]
            y_train, y_val = y[train_index], y[val_index]
        else:
            x_train = x
            y_train = y

        # use generic score function for any kind of purpose
        if generic_score:

            score[split_index] = generic_score(**locals())

        else:

            if not score:
                score = {"train_cv": [], "val_cv": []}

            # load learner
            try:
                learner = copy.deepcopy(model)
            except TypeError:
                learner = model
            learner.set_random_seed(seed, split_index, random_model)

            # check if model has already been computed
            learner.load_model()

            # fit learner
            learner.fit(x_train, y_train)

            # use custom score function for machine learning models having "fit" and "predict" methods
            if score_fun:

                # predict
                y_train_pred = learner.predict(x_train)
                y_val_pred = learner.predict(x_val)

                # compute score
                score_train = score_fun(y_train, y_train_pred)
                score_val = score_fun(y_val, y_val_pred)

                y_pred_list.append(y_val_pred)
                y_list.append(y_val)

            # use default score function provided by the model
            else:
                # compute score directly
                score_train = learner.score(x_train, y_train)
                score_val = learner.score(x_val, y_val)

                y_pred_list.append(learner.predict(x_val))
                y_list.append(y_val)

            # save results
            score["train_cv"].append(score_train)
            score["val_cv"].append(score_val)

            log_info.info("\tScore: train %.4f - validation %.4f" % (score_train, score_val))

            # save trained model
            learner.save_model()
            fitted_models.append(learner)

        split_index += 1

    return score, fitted_models, y_pred_list, y_list


def _compute_result_summary(models: List[Wrapper], model_id_list: List[List],
                            random_data: bool, random_model: bool,
                            seed: int, n_splits: int, scoring: [Callable, str],
                            test: Callable, alpha: int, cl: float,
                            train_cv: list, val_cv: list,
                            pvalues: list, best_solutions: list) -> pd.DataFrame:
    """
    Compute a summary of the cross-validation and model comparison results.

    Parameters
    --------
    :param models: list of machine learning models (keras or sklearn)
    :param model_id_list: list of model identifiers
    :param random_data: if True it enables data randomization
    :param random_model: if True it enables model randomization (if applicable)
    :param seed: seed used to make results reproducible
    :param n_splits: number of cross-validation iterations
    :param scoring: metric used to evaluate the model performance (f1 or accuracy)
    :param test: statistical test
    :param alpha: significance level
    :param cl: confidence level
    :param train_cv: cross-validation train scores
    :param val_cv: cross-validation validation scores
    :param pvalues: p-values of the statistical hypothesis test
    :param best_solutions: best solutions' indexes
    :return: summary of the results as a table
    """

    columns = [
        "db-name", "db-did",
        "model_name", "model_id", "module", "version", "parameters", "fit_params", "submodels", "is_standalone",
        "train_cv", "val_cv",
        "mean", "ci-l-bound", "ci-u-bound", "separable", "pvalue",
        "test", "alpha", "metric",
        "random-data", "random-model", "seed", "n-splits",
    ]

    results = pd.DataFrame(columns=columns)

    base_row = [
        test.__name__,
        alpha,
        str(scoring),

        random_data,
        random_model,
        seed,
        n_splits,
    ]

    index = 0
    for model in models:
        # compute confidence intervals of the mean of the validation score
        ci_bounds = confidence_interval_mean_t(val_cv[index], cl=cl)

        separable = False if index in best_solutions else True

        try:
            model_names = [str(m.model_name) for m in model.models]
        except:
            model_names = ""

        row = [
            model.dataset_name,
            model.dataset_id,

            model.model_name,
            model_id_list[index],
            model.model_type,
            model.version,
            model.parameters,
            str(model.fit_parameters),
            model_names,
            str(model.is_standalone),
            train_cv[index],
            val_cv[index],
            np.mean(val_cv[index]),
            ci_bounds[0],
            ci_bounds[1],
            separable,
            pvalues[index],
        ]
        row.extend(base_row)

        index += 1

        results = results.append(pd.DataFrame([row], columns=columns), ignore_index=True)

    return results


def _generate_detailed_summary(models: List[Wrapper], confusion_matrix_list: List[pycm.ConfusionMatrix],
                               best_indexes: List[int], output_dir: str) -> None:
    """
    Generate detailed summary about classification results, for the best models only.

    Parameters
    --------
    :param models: list of fitted models
    :param confusion_matrix_list:
    :param best_indexes: indexes of the best models
    :param output_dir: directory where results will be saved
    :return: None
    """

    summary_overall = pd.DataFrame()
    summary_class = pd.DataFrame()

    for i in best_indexes:

        model = models[i]
        confusion_matrix = confusion_matrix_list[i]

        if confusion_matrix:

            if model.model_id:
                model_id = model.model_name + " " + str(model.model_id)
            else:
                model_id = " ".join([m.model_name for m in model.models])

            # overall summary
            prefix_list = []
            for c in confusion_matrix.classes:
                prefix_list.append([model_id, "overall", c])
            prefix_overall = pd.DataFrame(prefix_list, columns=["model", "summary type", "classes"])

            overall_stat = pd.DataFrame.from_dict(confusion_matrix.overall_stat)
            summary = pd.concat([prefix_overall, overall_stat], axis=1).reset_index().drop(columns=["index"])

            summary_overall = pd.concat([summary_overall, summary], ignore_index=True)

            # class summary
            prefix_list = []
            for c in confusion_matrix.classes:
                prefix_list.append([model_id, "class summary", c])
            prefix_class = pd.DataFrame(prefix_list, columns=["model", "summary type", "classes"])

            class_stat = pd.DataFrame.from_dict(confusion_matrix.class_stat).reset_index()
            summary = pd.concat([prefix_class, class_stat], axis=1).reset_index().drop(columns=["index"])

            summary_class = pd.concat([summary_class, summary], ignore_index=True)

    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    summary_class.to_csv(os.path.join(output_dir, "summary_class.csv"))
    summary_overall.to_csv(os.path.join(output_dir, "summary_overall.csv"))


def compare_models(models: List[Wrapper],
                   x_train: np.ndarray, y_train: np.ndarray,
                   x_val: np.ndarray = None, y_val: np.ndarray = None,
                   random_data: bool = True, random_model: bool = True,
                   seed: int = 42, n_splits: int = 10, score_fun: Callable = None,
                   generic_score: Callable = None,
                   test: Callable = mannwhitneyu, alpha: int = 0.05, cl: float = 0.05,
                   experiment_name: str = "model_comparison", output_dir: str = "./output",
                   verbose: bool = False, class_names: dict = None, font_scale: float = 1,
                   encoding: str = "categorical",
                   bar_position: int = 0) -> pd.DataFrame:
    """
    Compare machine learning models' performance on the provided data set, using
    cross-validation and statistical hypothesis tests.

    Examples
    --------
    >>> from sklearn.linear_model import LogisticRegression, RidgeClassifier
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from sklearn.datasets import make_classification
    >>> import pandas as pd
    >>> import lazygrid as lg
    >>>
    >>> x, y = make_classification(random_state=42)
    >>>
    >>> lg_model_1 = lg.wrapper.SklearnWrapper(LogisticRegression())
    >>> lg_model_2 = lg.wrapper.SklearnWrapper(RandomForestClassifier())
    >>> lg_model_3 = lg.wrapper.SklearnWrapper(RidgeClassifier())
    >>>
    >>> models = [lg_model_1, lg_model_2, lg_model_3]
    >>> results = lg.model_selection.compare_models(models=models, x_train=x, y_train=y)
    >>>
    >>> pd.set_option('display.width', 9)
    >>> results[['model_name', 'module', 'version', 'ci-l-bound', 'ci-u-bound', 'pvalue']] #doctest: +ELLIPSIS
                   model_name   module version  ci-l-bound  ci-u-bound    pvalue
    0      LogisticRegression  sklearn  0.21.2    0.424323    0.585103  0.000123
    1  RandomForestClassifier  sklearn  0.21.2    0.851488    1.000000  0.829392
    2         RidgeClassifier  sklearn  0.21.2    0.896696    1.000000  1.000000

    Parameters
    --------
    :param models: list of machine learning models (keras or sklearn)
    :param x_train: training data
    :param y_train: input labels
    :param x_val: validation data
    :param y_val: validation labels
    :param random_data: if True it enables data randomization
    :param random_model: if True it enables model randomization (if applicable)
    :param seed: seed used to make results reproducible
    :param n_splits: number of cross-validation iterations
    :param score_fun: sklearn-like scoring function used to evaluate the model performance
    :param generic_score: generic score function
    :param test: statistical test
    :param alpha: significance level
    :param cl: confidence level
    :param experiment_name: name of the current experiment
    :param output_dir: path to the folder where the results will be saved (as csv file)
    :param verbose: if True enables plots
    :param class_names: dictionary of label names like {0: "Class 1", 1: "Class 2"}
    :param font_scale: font size of figures
    :param encoding: label encoding; accepted values are: "categorical" or "one-hot"
    :param bar_position: progress bar position
    :return: Pandas DataFrame containing a summary of the results
    """

    confusion_matrix_list = []
    train_cv = []
    val_cv = []
    models_id_list = []

    # Cross-validation
    i = 0
    progress_bar = tqdm(models, desc="Comparing models", position=bar_position, leave=True)
    for model in progress_bar:

        score, fitted_models, y_pred_list, y_true_list = cross_validation(model, x=x_train, y=y_train,
                                                                          x_val=x_val, y_val=y_val,
                                                                          random_data=random_data,
                                                                          random_model=random_model,
                                                                          seed=seed, n_splits=n_splits,
                                                                          score_fun=score_fun,
                                                                          generic_score=generic_score,
                                                                          bar_position=bar_position+1,
                                                                          verbose=verbose)

        conf_mat = generate_confusion_matrix(fitted_models[-1].model_id, fitted_models[-1].model_name,
                                             y_pred_list, y_true_list, class_names,
                                             font_scale, output_dir, encoding)

        confusion_matrix_list.append(conf_mat)
        models_id_list.append([m.model_id for m in fitted_models])
        models[i] = fitted_models[-1]
        train_cv.append(score["train_cv"])
        val_cv.append(score["val_cv"])
        i += 1

    # Find best solution
    best_sol, best_solutions, pvalues = find_best_solution(val_cv,
                                                           test=mannwhitneyu, alpha=alpha,
                                                           use_continuity=False, alternative="two-sided")

    # Compute results' summary
    results = _compute_result_summary(models, models_id_list, random_data, random_model,
                                      seed, n_splits, score_fun, test, alpha, cl,
                                      train_cv, val_cv, pvalues, best_solutions)

    # Save results to csv
    experiment = experiment_name + ".csv"
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)
    results.to_csv(os.path.join(output_dir, experiment))

    # save detailed summary for the best models
    _generate_detailed_summary(models, confusion_matrix_list, best_solutions, output_dir)

    return results



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
import glob
import logging
import traceback
from difflib import SequenceMatcher
import numpy as np
import sys
import openml
import pandas as pd
import os
from logging import Logger
from sklearn.impute import SimpleImputer
from sklearn.datasets import fetch_openml
from sklearn.preprocessing import LabelEncoder
from .logger import log_warn, log_info


def load_npy_dataset(path_x: str, path_y: str) -> (np.ndarray, np.ndarray, int):
    """
    Load npy data set.

    Examples
    --------
    >>> import os
    >>> from sklearn.datasets import make_classification
    >>> import numpy as np
    >>> import lazygrid as lg
    >>>
    >>> x, y = make_classification(random_state=42)
    >>>
    >>> path_x, path_y = "x.npy", "y.npy"
    >>> np.save(path_x, x)
    >>> np.save(path_y, y)
    >>>
    >>> x, y, n_classes = lg.datasets.load_npy_dataset(path_x, path_y)

    Parameters
    --------
    :param path_x: path to data matrix
    :param path_y: path to data labels
    :return: data matrix, data labels, and number of classes
    """

    try:
        x = np.load(path_x)
        y = np.load(path_y)
        n_classes = len(np.unique(y))

    except FileNotFoundError:
        log_warn.exception("Exception occurred")
        return None, None, None

    return x, y, n_classes


def load_openml_dataset(data_id: int = None, dataset_name: str = None) -> (np.ndarray, np.ndarray, int):
    """
    Load OpenML data set.

    Examples
    --------

    >>> import lazygrid as lg
    >>>
    >>> x, y, n_classes = lg.datasets.load_openml_dataset(dataset_name="iris")
    >>> n_classes
    3

    Parameters
    --------
    :param data_id: data set identifier
    :param dataset_name: data set name
    :return: data matrix, data labels, and number of classes
    """

    assert (data_id is not None) or (dataset_name is not None)
    assert isinstance(dataset_name, str) or dataset_name is None

    try:
        if data_id is not None:
            x, y = fetch_openml(data_id=data_id, return_X_y=True)
        else:
            x, y = fetch_openml(name=dataset_name, return_X_y=True)

        if not isinstance(x, np.ndarray):
            x = x.toarray()

        si = SimpleImputer(missing_values=np.nan, strategy='mean')
        x = si.fit_transform(x)

        le = LabelEncoder()
        y = le.fit_transform(y)
        n_classes = np.unique(y).shape[0]

        return x, y, n_classes

    except Exception:
        log_warn.exception("Exception occurred")
        return [None, None, None]


def _similar(a: str, b: str) -> float:
    """
    Compute how much two strings are similar.
    """
    return SequenceMatcher(None, a, b).ratio()


def _is_correct_task(task: str, db: dict) -> bool:
    """
    Check if the current data set is compatible with the specified task.

    Parameters
    --------
    :param task: regression or classification
    :param db: OpenML data set dictionary
    :return: True if the task and the data set are compatible
    """
    if task == "classification":
        return db['NumberOfSymbolicFeatures'] == 1 and db['NumberOfClasses'] > 0
    elif task == "regression":
        return True
    else:
        return False


def _load_datasets(output_dir: str = "./data", min_classes: int = 0, task: str = "classification",
                   max_samples: int = np.inf, max_features: int = np.inf) -> pd.DataFrame:
    """
    Load all OpenML data sets compatible with the requirements and save a .csv file as a reference.

    Examples
    --------
    >>> datasets = _load_datasets(task="classification", min_classes=2, max_samples=1000, max_features=10)
    >>> datasets.loc["iris"]
    version          45
    did           42098
    n_samples       150
    n_features        4
    n_classes         3
    Name: iris, dtype: int64

    Parameters
    --------
    :param output_dir: directory where the .csv file will be stored
    :param min_classes: minimum number of classes required for each data set
    :param task: classification or regression
    :param max_samples: maximum number of samples required for each data set
    :param max_features: maximum number of features required for each data set
    :return: dataframe containing the information required to load the latest version of each data set
    """

    data = {}

    for key, db in openml.datasets.list_datasets().items():

        try:
            log_info.info("Loading data set: %s, ID: %d..." % (db['name'], db['did']))

            if db['NumberOfClasses'] > min_classes and _is_correct_task(task, db) and \
                    db['NumberOfInstances'] < max_samples and db['NumberOfFeatures'] < max_features and \
                    db['status'] == 'active':

                if db['name'] in data:
                    if db['version'] < data[db['name']]['version']:
                        continue

                # load_openml_dataset(db['did'])

                data[db['name']] = {}
                data[db['name']]['version'] = db['version']
                data[db['name']]['did'] = db['did']
                data[db['name']]['n_samples'] = db['NumberOfInstances']
                data[db['name']]['n_features'] = db['NumberOfNumericFeatures']
                data[db['name']]['n_classes'] = db['NumberOfClasses']

        except (IndexError, ValueError, KeyError):
            log_info.info("Error loading data set: %s, ID: %d!" % (db['name'], db['did']))
            log_info.info(traceback.format_exc())

    data = pd.DataFrame(data).transpose()
    try:
        data = data.sort_values(by=["n_samples", "n_features", "n_classes"]).astype('int64')
    except KeyError:
        log_info.info(traceback.format_exc())
        return None

    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)
    file_name = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S") + "-datalist.csv"
    data.to_csv(os.path.join(output_dir, file_name), index_label="db_name")

    return data


def fetch_datasets(output_dir: str = "./data", update_data: bool = False,
                   min_classes: int = 0, task: str = "classification",
                   max_samples: int = np.inf, max_features: int = np.inf) -> pd.DataFrame:
    """
    Load OpenML data sets compatible with the requirements.

    Examples
    --------
    >>> import lazygrid as lg
    >>>
    >>> datasets = lg.datasets.fetch_datasets(task="classification", min_classes=2, max_samples=1000, max_features=10)
    >>> datasets.loc["iris"]
    version          45
    did           42098
    n_samples       150
    n_features        4
    n_classes         3
    Name: iris, dtype: int64

    Parameters
    --------
    :param output_dir: directory where the .csv file will be stored
    :param update_data: if True it deletes cached data sets and downloads their latest version;
                        otherwise it loads data sets as specified inside the cache
    :param min_classes: minimum number of classes required for each data set
    :param task: classification or regression
    :param max_samples: maximum number of samples required for each data set
    :param max_features: maximum number of features required for each data set
    :return: dataframe containing the information required to load the latest version of each data set
    """

    files_location = os.path.join(output_dir, '*.csv')
    file_list = glob.glob(files_location)

    # download (again) new data if necessary
    if not os.path.isdir(output_dir) or not file_list or update_data:
        data = _load_datasets(output_dir, min_classes, task, max_samples, max_features)

        # delete previous data files
        file_list.sort()
        for file in file_list[:-1]:
            os.remove(file)

    # fetch last version of downloaded data if there is one
    elif file_list:
        file_list = glob.glob(files_location)
        file_list.sort()
        data = pd.read_csv(file_list[-1], index_col=0).astype("int64")

    else:
        log_warn.exception("Exception occurred")
        raise ValueError("Data not found")

    return data



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Collection

from keras import Sequential
from keras.layers import Dense
from keras.initializers import glorot_uniform  # Or your initializer of choice
from keras import optimizers
import keras.backend as K


def reset_weights(model: Sequential, seed: int = 42):
    """
    Random reset of neural network's weights.

    Examples
    --------
    >>> from keras import Sequential
    >>> import numpy as np
    >>> import random
    >>> import tensorflow as tf
    >>> from keras import backend as K
    >>> import os
    >>> import lazygrid as lg
    >>>
    >>> # These lines are mandatory if you need to get the same results each time you use the network!
    >>> seed = 42
    >>> np.random.seed(seed)
    >>> random.seed(seed)
    >>> os.environ['PYTHONHASHSEED'] = str(seed)
    >>> config = tf.ConfigProto(intra_op_parallelism_threads=1, inter_op_parallelism_threads=1)
    >>> tf.set_random_seed(seed)
    >>> sess = tf.Session(graph=tf.get_default_graph(), config=config)
    >>> K.set_session(sess)
    >>>
    >>> keras_model = lg.neural_models.keras_classifier(layers=[10, 5], input_shape=(20,), n_classes=2)
    >>> lg.neural_models.reset_weights(keras_model, seed=seed)

    Parameters
    --------
    :param model: Keras, theano, or tensorflow model
    :param seed: seed used to make results reproducible
    :return: None
    """

    # 1. Set `PYTHONHASHSEED` environment variable at a fixed value
    import os
    os.environ['PYTHONHASHSEED'] = str(seed)

    # 2. Set `python` built-in pseudo-random generator at a fixed value
    import random
    random.seed(seed)

    # 3. Set `numpy` pseudo-random generator at a fixed value
    import numpy as np
    np.random.seed(seed)

    # 4. Set `tensorflow` pseudo-random generator at a fixed value
    import tensorflow as tf
    tf.set_random_seed(seed)

    # get old weights
    initial_weights = model.get_weights()

    # check backend
    backend_name = K.backend()
    if backend_name == 'tensorflow':
        # # 5. Configure a new global `tensorflow` session
        # from keras import backend as K
        # session_conf = tf.ConfigProto(intra_op_parallelism_threads=1, inter_op_parallelism_threads=1)
        # sess = tf.Session(graph=tf.get_default_graph(), config=session_conf)
        # K.set_session(sess)
        k_eval = lambda placeholder: placeholder.eval(session=K.get_session())
    elif backend_name == 'theano':
        k_eval = lambda placeholder: placeholder.eval()
    else:
        raise ValueError("Unsupported backend")

    # set new weights only if they are trainable
    new_weights = []
    for w, l in zip(initial_weights, model.layers):
        if l.trainable:
            new_weights.append(k_eval(glorot_uniform(seed=seed)(w.shape)))
        else:
            new_weights.append(w)
    model.set_weights(new_weights)


def keras_classifier(layers: list, input_shape: tuple, n_classes: int,
                     verbose: bool = True, metrics: Collection = ["accuracy"],
                     model_name: str = "MyKerasNet", lr: float = 0.1) -> Sequential:
    """
    Generate keras feed-forward neural model for classification.

    Examples
    --------
    >>> from keras import Sequential
    >>> import lazygrid as lg
    >>>
    >>> keras_model = lg.neural_models.keras_classifier(layers=[10, 5], input_shape=(20,), n_classes=2)

    Parameters
    --------
    :param layers: list of layers
    :param input_shape: shape of input data
    :param n_classes: number of classes
    :param verbose: if True it prints the summary of the model
    :param metrics: metrics used by the optimizer
    :param model_name: name of the keras network
    :param lr: learning rate
    :return: compiled keras model
    """

    assert n_classes >= 2

    output_size = 1 if n_classes == 2 else n_classes
    if n_classes > 2:
        activation = "softmax"
        loss = "categorical_crossentropy"
    else:
        activation = "sigmoid"
        loss = "binary_crossentropy"

    model = Sequential()
    model.add(Dense(layers[0], activation="relu", input_shape=input_shape))
    for layer in layers[1:]:
        model.add(Dense(layer, activation="relu"))
    model.add(Dense(output_size, activation=activation))

    model.__name__ = model_name

    if verbose:
        model.summary()

    model.compile(loss=loss,
                  optimizer=optimizers.Adam(lr=lr),
                  metrics=metrics)

    return model



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sqlite3
from typing import Optional, Any, Iterable
import numpy as np
from .logger import log_warn


def save_to_db(db_name: str, entry: Iterable, query: Iterable,
               create_stmt: str, insert_stmt: str, query_stmt: str) -> Optional[Any]:
    """
    Save fitted model into a database.

    Examples
    --------
    >>> import lazygrid as lg
    >>> from sklearn.linear_model import LogisticRegression
    >>>
    >>> model = lg.wrapper.SklearnWrapper(LogisticRegression())
    >>>
    >>> entry = model.get_entry()
    >>> query = model.get_query()
    >>>
    >>> query_result = lg.database.save_to_db(model.db_name, entry, query)
    >>>
    >>> db_entry = lg.database.load_from_db(model.db_name, query)

    Parameters
    --------
    :param db_name: database name
    :param entry: collection that will be inserted into the database
    :param query: collection required to fetch a database entry
    :param create_stmt: database statement for table creation
    :param insert_stmt: database statement for model insertion
    :param query_stmt: database statement for model query
    :return: query result
    """

    # Sqlite does not accept INT larger than 8 bytes.
    sqlite3.register_adapter(np.int64, lambda val: int(val))
    sqlite3.register_adapter(np.int32, lambda val: int(val))

    # Create database if does not exists
    db_dir = "./database"
    if not os.path.isdir(db_dir):
        os.mkdir(db_dir)
    db_name = os.path.join(db_dir, db_name + ".sqlite")
    db = sqlite3.connect(db_name)
    cursor = db.cursor()
    db.execute(create_stmt)

    # Insert item
    try:
        cursor.execute(insert_stmt, entry)
    except sqlite3.IntegrityError:
        log_warn.exception("Exception occurred")
        pass
    result = cursor.execute(query_stmt, query).fetchone()

    db.commit()
    db.close()

    return result


def load_from_db(db_name: str, query: Iterable, create_stmt: str, query_stmt: str) -> Optional[Any]:
    """
    Load fitted model from a database.

    Examples
    --------
    >>> import lazygrid as lg
    >>> from sklearn.linear_model import LogisticRegression
    >>>
    >>> model = lg.wrapper.SklearnWrapper(LogisticRegression())
    >>>
    >>> entry = model.get_entry()
    >>> query = model.get_query()
    >>>
    >>> query_result = lg.database.save_to_db(model.db_name, entry, query)
    >>>
    >>> db_entry = lg.database.load_from_db(model.db_name, query)

    Parameters
    --------
    :param db_name: database name
    :param query: collection required to fetch a database entry
    :param create_stmt: database statement for table creation
    :param query_stmt: database statement for model query
    :return: query result
    """

    # Sqlite does not accept INT larger than 8 bytes.
    sqlite3.register_adapter(np.int64, lambda val: int(val))
    sqlite3.register_adapter(np.int32, lambda val: int(val))

    # Connect to database if it exists
    db_dir = "./database"
    if not os.path.isdir(db_dir):
        os.mkdir(db_dir)
    db_name = os.path.join(db_dir, db_name + ".sqlite")
    db = sqlite3.connect(db_name)
    cursor = db.cursor()
    db.execute(create_stmt)

    result = cursor.execute(query_stmt, query).fetchone()
    cursor.close()

    return result


def load_all_from_db(db_name: str, table_name: str = "MODEL") -> Optional[Any]:
    """
    Load all database items.

    Examples
    --------
    >>> from sklearn.linear_model import RidgeClassifier
    >>> from sklearn.datasets import make_classification
    >>> import lazygrid as lg
    >>>
    >>> x, y = make_classification(random_state=42)
    >>> model = lg.wrapper.SklearnWrapper(RidgeClassifier(),
    ...                                   db_name=db_name, dataset_id=1,
    ...                                   dataset_name="make-classification")
    >>> scores, _, _, _ = lg.model_selection.cross_validation(model, x, y)
    >>>
    >>> db_entries = lg.database.load_all_from_db(db_name)

    Parameters
    --------
    :param db_name: database name
    :param table_name: name of table to be loaded
    :return: query result
    """

    # Sqlite does not accept INT larger than 8 bytes.
    sqlite3.register_adapter(np.int64, lambda val: int(val))
    sqlite3.register_adapter(np.int32, lambda val: int(val))

    # Connect to database if it exists
    db_dir = "./database"
    if not os.path.isdir(db_dir):
        os.mkdir(db_dir)
    db_name = os.path.join(db_dir, db_name + ".sqlite")
    db = sqlite3.connect(db_name)
    cursor = db.cursor()

    stmt = '''SELECT name FROM sqlite_master WHERE type='table' AND name=? '''
    table = cursor.execute(stmt, (table_name,)).fetchone()
    if not table:
        db.close()
        return None

    query_stmt = '''SELECT * FROM %s''' % table_name
    result = cursor.execute(query_stmt).fetchall()
    cursor.close()

    return result


def drop_db(db_name: str) -> None:
    """
    Drop database table if it exists.

    Examples
    --------
    >>> import lazygrid as lg
    >>>
    >>> lg.database.drop_db(db_name="my-database")

    Parameters
    --------
    :param db_name: database name
    :return: None
    """

    db_dir = "./database"
    if not os.path.isdir(db_dir):
        os.mkdir(db_dir)
    db_name = os.path.join(db_dir, db_name + ".sqlite")

    db = sqlite3.connect(db_name)
    cursor = db.cursor()

    cursor.execute("DROP TABLE IF EXISTS MODEL")

    db.close()

    return




# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import matplotlib
matplotlib.use('Agg')
import os
from typing import List
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import pycm


def plot_confusion_matrix(confusion_matrix: pd.DataFrame, font_scale: float,
                          file_name: str, title: str = 'Confusion matrix') -> None:
    """
    Generate confusion matrix figure.

    >>> import pandas as pd
    >>> import lazygrid as lg
    >>>
    >>> conf_mat_dict = {"N": {"N": 48, "P": 2},
    ...                  "P": {"N": 5, "P": 45}}
    >>> confusion_matrix = pd.DataFrame.from_dict(conf_mat_dict)
    >>>
    >>> lg.plotter.plot_confusion_matrix(confusion_matrix, font_scale=1, file_name="conf_mat.png")

    Parameters
    --------
    :param confusion_matrix: confusion matrix dataframe
    :param font_scale: font size
    :param file_name: figure file name
    :param title: figure title
    :return: None
    """

    sns.set(font_scale=font_scale)

    plt.figure()
    ax = sns.heatmap(confusion_matrix, annot=True, vmin=0, linewidths=.3, cmap="Greens", square=True, fmt='d')
    ax.set(xlabel='Prediction', ylabel='True', title=title)
    plt.savefig(file_name, dpi=800)
    plt.close()

    return


def one_hot_list_to_categorical(y_one_hot_list: List[np.ndarray]) -> np.ndarray:
    """
    Transform list of one-hot-encoded labels into a categorical array of labels.

    Examples
    --------
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from sklearn.datasets import make_classification
    >>> import lazygrid as lg
    >>>
    >>> x, y = make_classification(random_state=42)
    >>>
    >>> classifier = RandomForestClassifier(random_state=42)
    >>>
    >>> model = lg.wrapper.SklearnWrapper(classifier)
    >>> score, fitted_models, y_pred_list, y_true_list = lg.model_selection.cross_validation(model=model, x=x, y=y)
    >>>
    >>> y_true_categorical = lg.plotter.one_hot_list_to_categorical(y_true_list)

    Parameters
    --------
    :param y_one_hot_list: one-hot-encoded list of labels
    :return: categorical array of labels
    """
    y_categorical_list = []
    for y_one_hot in y_one_hot_list:
        y_categorical_list.append(np.argmax(y_one_hot, axis=1))
    return np.hstack(y_categorical_list)


def generate_confusion_matrix(model_id: int, model_name: str,
                              y_pred_list: List[np.ndarray], y_true_list: List[np.ndarray],
                              class_names: dict = None, font_scale: float = 1,
                              output_dir: str = "./figures",
                              encoding: str = "categorical") -> pycm.ConfusionMatrix:
    """
    Generate and save confusion matrix.

    Examples
    --------
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from sklearn.datasets import make_classification
    >>> import lazygrid as lg
    >>>
    >>> x, y = make_classification(random_state=42)
    >>>
    >>> classifier = RandomForestClassifier(random_state=42)
    >>>
    >>> model = lg.wrapper.SklearnWrapper(classifier)
    >>> score, fitted_models, y_pred_list, y_true_list = lg.model_selection.cross_validation(model=model, x=x, y=y)
    >>>
    >>> conf_mat = lg.plotter.generate_confusion_matrix(fitted_models[-1].model_id, fitted_models[-1].model_name,
    ...                                                 y_pred_list, y_true_list, class_names={0: "N", 1: "P"})

    Parameters
    --------
    :param model_id: model identifier
    :param model_name: model name
    :param y_pred_list: predicted labels list
    :param y_true_list: true labels list
    :param class_names: dictionary of label names like {0: "Class 1", 1: "Class 2"}
    :param font_scale: figure font size
    :param output_dir: output directory
    :param encoding: kind of label encoding
    :return: confusion matrix object
    """

    # transform labels
    if encoding == "categorical":
        y_pred = np.hstack(y_pred_list)
        y_true = np.hstack(y_true_list)
    elif encoding == "one-hot":
        y_pred = one_hot_list_to_categorical(y_pred_list)
        y_true = one_hot_list_to_categorical(y_true_list)
    else:
        return None

    conf_mat = pycm.ConfusionMatrix(actual_vector=y_true, predict_vector=y_pred)

    # rename classes
    if class_names:
        conf_mat.relabel(mapping=class_names)

    conf_mat_pd = pd.DataFrame.from_dict(conf_mat.matrix).T

    # figure title and file name
    name = model_name + "_" + str(model_id)
    title = model_name + " " + str(model_id)
    file_name = os.path.join(output_dir, "conf_mat_" + name + ".png")
    title = title
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    plot_confusion_matrix(conf_mat_pd, font_scale, file_name, title)

    return conf_mat


def plot_boxplots(scores: List, labels: List[str], file_name: str, title: str, output_dir: str = "./figures") -> dict:
    """
    Generate and save boxplots.

    Examples
    --------
    >>> from sklearn.linear_model import LogisticRegression, RidgeClassifier
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from sklearn.datasets import make_classification
    >>> import lazygrid as lg
    >>>
    >>> x, y = make_classification(random_state=42)
    >>>
    >>> lg_model_1 = lg.wrapper.SklearnWrapper(LogisticRegression())
    >>> lg_model_2 = lg.wrapper.SklearnWrapper(RandomForestClassifier())
    >>> lg_model_3 = lg.wrapper.SklearnWrapper(RidgeClassifier())
    >>>
    >>> models = [lg_model_1, lg_model_2, lg_model_3]
    >>>
    >>> score_list = []
    >>> labels = []
    >>> for model in models:
    >>>     scores, _, _, _ = lg.model_selection.cross_validation(model, x, y)
    >>>     score_list.append(scores["val_cv"])
    >>>     labels.append(model.model_name)
    >>>
    >>> file_name = "val_scores"
    >>> title = "Model comparison"
    >>> lg.plotter.plot_boxplots(score_list, labels, file_name, title)

    Parameters
    --------
    :param scores: list of scores to compare
    :param labels: name / identifier of each score list
    :param file_name: output file name
    :param title: figure title
    :param output_dir: output directory
    :return: boxplot object
    """
    file_name = os.path.join(output_dir, "box_plot_" + file_name + ".png")
    if not os.path.isdir(output_dir):
        os.mkdir(output_dir)

    cv = np.stack(scores, axis=1)

    plt.figure()
    results = plt.boxplot(cv, notch=True, labels=labels)
    plt.title(title)
    plt.tight_layout()
    plt.savefig(file_name, dpi=800)
    plt.show()

    return results



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='lazygrid',
    version='4.0.0',
    description='LazyGrid: Automatic, efficient and flexible implementation of complex machine '
                'learning pipeline generation and cross-validation.',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url='https://github.com/glubbdubdrib/lazygrid',
    author='Pietro Barbiero and Giovanni Squillero',
    author_email='cleisthenes.megacleos@gmail.com',
    license="Apache 2.0",
    packages=find_packages(exclude=('tests', 'docs')),
    include_package_data=True,
)



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

create_model_stmt = '''CREATE TABLE IF NOT EXISTS MODEL(
        id INTEGER PRIMARY KEY,
        name TEXT NOT NULL,
        type TEXT NOT NULL,
        class BLOB NOT NULL,
        parameters TEXT NOT NULL,
        fit_parameters TEXT NOT NULL,
        predict_parameters TEXT NOT NULL,
        score_parameters TEXT NOT NULL,
        version TEXT NOT NULL,
        is_standalone INTEGER NOT NULL,
        models_id TEXT NOT NULL,
        dataset_id INTEGER NOT NULL,
        dataset_name TEXT,
        cv_split INTEGER NOT NULL,
        previous_step_id INTEGER,
        serialized_model BLOB NOT NULL,
        UNIQUE (name, type, parameters, 
                fit_parameters, predict_parameters, score_parameters,
                version, models_id, dataset_id, 
                cv_split, previous_step_id)
        )'''

insert_model_stmt = '''INSERT INTO MODEL(
        name, type, class, parameters, 
        fit_parameters, predict_parameters, score_parameters, 
        version, is_standalone, models_id,
        dataset_id, dataset_name, cv_split, previous_step_id, serialized_model)
        VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)'''

query_model_stmt = '''SELECT id, type, class, serialized_model, fit_parameters, is_standalone FROM MODEL
                      WHERE name=? AND type=? AND parameters=? AND 
                            fit_parameters=? AND predict_parameters=? AND score_parameters=? AND
                            version=? AND models_id=? AND dataset_id=? AND 
                            cv_split=? AND previous_step_id=?'''



# -*- coding: utf-8 -*-
#
# Copyright 2019 Pietro Barbiero and Giovanni Squillero
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License.
# You may obtain a copy of the License at:
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
import pickle
import sys
from typing import Callable, Any, Union
import json
import keras
import sklearn
from keras import Model
from sklearn.ensemble import RandomForestClassifier
from abc import ABC
import tensorflow as tf
from .database import drop_db, save_to_db, load_from_db
from .neural_models import reset_weights
from .config import create_model_stmt, insert_model_stmt, query_model_stmt


class Wrapper(ABC):
    """
    Abstract wrapper class for machine learning models.

    Examples
    --------
    >>> from lazygrid.wrapper import Wrapper
    >>>
    >>>
    >>> class CustomWrapper(Wrapper):
    >>>
    >>>     def __init__(self, **kwargs):
    >>>         Wrapper.__init__(self, **kwargs)
    >>>
    >>>     def set_random_seed(self, seed, split_index, random_model):
    >>>         pass
    >>>
    >>>     def parse_parameters(self) -> str:
    >>>         pass
    """

    def __init__(self, model: Any, dataset_id=None, dataset_name=None, db_name=None,
                 fit_params=None, predict_params=None, score_params=None, model_id=None,
                 cv_split=None, is_standalone=True, **kwargs):

        """
        Wrapper initialization.

        Parameters
        --------
        :param model: machine learning model
        :param dataset_id: data set identifier
        :param dataset_name: data set name
        :param db_name: database name
        :param fit_params: model's fit parameters
        :param predict_params: model's predict parameters
        :param score_params: model's score parameters
        :param model_id: model identifier
        :param cv_split: cross-validation split identifier
        :param is_standalone: True if model can be used independently from other models
        :param kwargs: other parameters
        """

        if not dataset_id:
            db_name = None
            dataset_id = 1
            dataset_name = "default-dataset"
        if not db_name:
            db_name = "default-db"
            drop_db(db_name)

        self.model_id = model_id
        self.model_name = str(model.__class__.__name__).split(".")[0]
        self.model_type = str(model.__module__).split(".")[0]
        self.version = sys.modules[self.model_type].__version__

        self.is_fitted = False
        self.is_standalone = is_standalone
        self.fit_parameters = dict_to_json(fit_params)
        self.predict_parameters = dict_to_json(predict_params)
        self.score_parameters = dict_to_json(score_params)
        self.models = None
        self.models_id = None

        self.model = model
        self.parameters = None
        self.serialized_model = None

        self.cv_split = cv_split
        self.dataset_id = dataset_id
        self.dataset_name = dataset_name
        self.db_name = db_name
        self.previous_step_id = -1

        self.entry = None
        self.query = None

    def get_entry(self) -> tuple:
        """
        Define database entry for the model.

        Parameters
        --------
        :return: database entry
        """

        entry = (
            self.model_name,
            self.model_type,
            pickle.dumps(self.__class__, protocol=2),
            str(self.parameters),
            self.fit_parameters,
            self.predict_parameters,
            self.score_parameters,
            self.version,
            int(self.is_standalone),
            str(self.models_id),
            self.dataset_id,
            self.dataset_name,
            self.cv_split,
            self.previous_step_id,
            self.serialized_model,
        )
        return entry

    def get_query(self) -> tuple:
        """
        Define database query for the model.

        Parameters
        --------
        :return: database query
        """
        query = (
            self.model_name,
            self.model_type,
            str(self.parameters),
            self.fit_parameters,
            self.predict_parameters,
            self.score_parameters,
            self.version,
            str(self.models_id),
            self.dataset_id,
            self.cv_split,
            self.previous_step_id,
        )
        return query

    def to_database(self, **kwargs):
        """
        Send model into a database.

        Parameters
        --------
        :return: query result
        """
        self.entry = self.get_entry()
        self.query = self.get_query()
        return save_to_db(self.db_name, self.entry, self.query, create_model_stmt, insert_model_stmt, query_model_stmt)

    def from_database(self, **kwargs):
        """
        Fetch model from a database.

        Parameters
        --------
        :return: query result
        """
        self.query = self.get_query()
        return load_from_db(self.db_name, self.query, create_model_stmt, query_model_stmt)

    def load_model(self, **kwargs) -> Any:
        """
        Load model from database if possible.

        Parameters
        --------
        :return: wrapped model
        """
        result = self.from_database()
        if result:
            model_id, model_type, model_class, serialized_model, fit_parameters, is_standalone = result
            self.is_fitted = True
            self.model_id = model_id
            self.model = pickle.loads(serialized_model)
            self.serialized_model = serialized_model
        return self

    def save_model(self, **kwargs) -> None:
        """
        Save model into database.

        Parameters
        --------
        :return: None
        """
        if not self.serialized_model:
            self.serialized_model = pickle.dumps(self.model, protocol=2)
        result = self.to_database()
        if result:
            model_id, model_type, model_class, serialized_model, fit_parameters, is_standalone = result
            self.model_id = model_id

    def set_random_seed(self, **kwargs):
        """
        Set model random state if possible.

        Parameters
        --------
        :return: None
        """
        raise NotImplementedError

    def parse_parameters(self, **kwargs) -> str:
        """
        Parse model parameters.

        Parameters
        --------
        :return: parameters as a string
        """
        raise NotImplementedError

    def fit(self, x, y, **kwargs) -> None:
        """
        Fit model with some samples.

        Parameters
        --------
        :param x: train data
        :param y: train labels
        :return: None
        """
        if not self.is_fitted:
            self.model.fit(x, y)
        self.is_fitted = True

    def predict(self, x, **kwargs) -> Any:
        """
        Predict labels for some input samples.

        Parameters
        --------
        :param x: input data
        :return: predictions
        """
        return self.model.predict(x)

    def score(self, x, y, **kwargs) -> Any:
        """
        Compute score for some input samples.

        Parameters
        --------
        :param x: input data
        :param y: input labels
        :return: score
        """
        return self.model.score(x, y)


class SklearnWrapper(Wrapper):
    """
    Class to wrap sklearn models.

    Examples
    --------
    >>> import lazygrid as lg
    >>> from sklearn.linear_model import LogisticRegression
    >>>
    >>> lg_model = lg.wrapper.SklearnWrapper(LogisticRegression())
    """

    def __init__(self, model: Any, dataset_id=None, dataset_name=None, db_name=None,
                 fit_params=None, predict_params=None, score_params=None, model_id=None,
                 cv_split=None, is_standalone=True):
        """
        Wrapper initialization.

        Parameters
        --------
        :param model: machine learning model
        :param dataset_id: data set identifier
        :param dataset_name: data set name
        :param db_name: database name
        :param fit_params: model's fit parameters
        :param predict_params: model's predict parameters
        :param score_params: model's score parameters
        :param model_id: model identifier
        :param cv_split: cross-validation split identifier
        :param is_standalone: True if model can be used independently from other models
        """
        model = corner_cases(model)
        Wrapper.__init__(self, model, dataset_id, dataset_name, db_name, fit_params,
                         predict_params, score_params, model_id, cv_split, is_standalone)
        self.parameters = self.parse_parameters()

    def set_random_seed(self, seed, split_index, random_model):
        """
        Set model random state if possible.

        Parameters
        --------
        :param seed: random seed
        :param split_index: cross-validation split identifier
        :param random_model: whether the model should have the same random state for each cross-validation split
        :return: None
        """
        if random_model:
            random_state = split_index
        else:
            random_state = seed
        try:
            self.model.set_params(**{"random_state": random_state})
        except (AttributeError, ValueError):
            pass
        self.cv_split = split_index
        self.parameters = self.parse_parameters()

    def parse_parameters(self) -> str:
        """
        Parse sklearn model parameters.

        Parameters
        --------
        :return: model parameters
        """
        return parse_sklearn_model(self.model)


class PipelineWrapper(Wrapper):
    """
    Class to wrap sklearn pipeline objects.

    Examples
    --------
    >>> import lazygrid as lg
    >>> from sklearn.ensemble import RandomForestClassifier
    >>> from sklearn.pipeline import Pipeline
    >>> from sklearn.preprocessing import StandardScaler
    >>> from sklearn.feature_selection import SelectKBest, mutual_info_classif
    >>>
    >>> standardizer = StandardScaler()
    >>> feature_selector = SelectKBest(score_func=mutual_info_classif, k=2)
    >>> classifier = RandomForestClassifier(random_state=42)
    >>> pipeline = Pipeline(steps=[("standardizer", standardizer),
    >>>                            ("feature_selector", feature_selector),
    >>>                            ("classifier", classifier)])
    >>>
    >>> lg_model = lg.wrapper.PipelineWrapper(pipeline)
    """

    def __init__(self, model, dataset_id=None, dataset_name=None, db_name=None,
                 fit_params=None, predict_params=None, score_params=None, model_id=None,
                 cv_split=None, is_standalone=True):
        """
        Wrapper initialization.

        Parameters
        --------
        :param model: machine learning model
        :param dataset_id: data set identifier
        :param dataset_name: data set name
        :param db_name: database name
        :param fit_params: model's fit parameters
        :param predict_params: model's predict parameters
        :param score_params: model's score parameters
        :param model_id: model identifier
        :param cv_split: cross-validation split identifier
        :param is_standalone: True if model can be used independently from other models
        """
        model = corner_cases(model)
        Wrapper.__init__(self, model, dataset_id, dataset_name, db_name, fit_params,
                         predict_params, score_params, model_id, cv_split, is_standalone)
        self.models = []
        self.models_id = []
        for step in model.steps:
            pipeline_step = SklearnWrapper(model=step[1], cv_split=cv_split, dataset_id=dataset_id,
                                           dataset_name=dataset_name, db_name=db_name, is_standalone=False)
            self.models.append(pipeline_step)
            self.models_id.append(pipeline_step.model_id)

        parameters = []
        for step in self.models:
            parameters.append(step.parameters)
        self.parameters = str(parameters)

    def save_model(self) -> None:
        """
        Save model into database.

        Parameters
        --------
        :return: None
        """
        self.models_id = []
        previous_step_id = -1
        # save each step separately
        for model in self.models:
            model.previous_step_id = previous_step_id
            model.save_model()
            self.models_id.append(model.model_id)
            previous_step_id = model.model_id
        # serialize model
        if not self.serialized_model:
            self.serialized_model = pickle.dumps(self.model, protocol=2)
        result = self.to_database()
        if result:
            model_id, model_type, model_class, serialized_model, fit_parameters, is_standalone = result
            self.model_id = model_id

    def load_model(self) -> Any:
        """
        Load model from database.

        Parameters
        --------
        :return: None
        """
        pipeline = []
        previous_step_id = -1
        i = 0
        is_fetchable = True
        # load each step separately (if present)
        for step in self.models:
            step.previous_step_id = previous_step_id
            if is_fetchable:
                fitted_step = step.load_model()
                if fitted_step.model_id and is_fetchable:
                    pipeline_step = ("id_" + str(fitted_step.model_id), fitted_step.model)
                    previous_step_id = fitted_step.model_id
                    self.models[i].is_fitted = True
                    self.models_id[i] = previous_step_id
                else:
                    # from now on the pipeline will be different, you can't fetch from database other models
                    is_fetchable = False
            if not is_fetchable:
                pipeline_step = ("n_" + str(i), copy.deepcopy(step.model))
            pipeline.append(pipeline_step)
            i += 1
        self.model.steps = pipeline
        result = self.from_database()
        if result:
            model_id, model_type, model_class, serialized_model, fit_parameters, is_standalone = result
            self.is_fitted = True
            self.model_id = model_id
            self.model = pickle.loads(serialized_model)
            self.serialized_model = serialized_model
        return self

    def set_random_seed(self, seed, split_index, random_model):
        """
        Set model random state if possible.

        Parameters
        --------
        :param seed: random seed
        :param split_index: cross-validation split identifier
        :param random_model: whether the model should have the same random state for each cross-validation split
        :return: None
        """
        if random_model:
            random_state = split_index
        else:
            random_state = seed
        for parameter in list(self.model.get_params().keys()):
            if "random_state" in parameter:
                self.model.set_params(**{parameter: random_state})
        self.cv_split = split_index
        # set random seed of pipeline steps
        for model in self.models:
            model.set_random_seed(seed, split_index, random_model)

    def parse_parameters(self, **kwargs) -> str:
        pass

    def fit(self, x_train, y_train, **kwargs):
        """
        Fit model with some samples.

        Parameters
        --------
        :param x_train: train data
        :param y_train: train labels
        :return: None
        """
        x_train_t = x_train
        i = 0
        for pipeline_step, model in zip(self.model.steps, self.models):
            if not model.is_fitted:
                # print("NOT FITTED!")
                pipeline_step[1].fit(x_train_t, y_train, )
                self.models[i] = SklearnWrapper(model=copy.deepcopy(pipeline_step[1]), cv_split=self.cv_split,
                                                dataset_id=self.dataset_id, dataset_name=self.dataset_name,
                                                db_name=self.db_name, is_standalone=False)
                self.models[i].is_fitted = True
            if hasattr(pipeline_step[1], "transform"):
                x_train_t = pipeline_step[1].transform(x_train_t)
            i += 1
        self.is_fitted = True


class KerasWrapper(Wrapper):
    """
    Class to wrap keras objects.

    Examples
    --------
    >>> import lazygrid as lg
    >>>
    >>> keras_nn = lg.neural_models.keras_classifier(layers=[10, 5], input_shape=(20,), n_classes=2)
    >>> lg_model = lg.wrapper.KerasWrapper(keras_nn)
    """

    def __init__(self, model: Any, dataset_id=None, dataset_name=None, db_name=None,
                 fit_params=None, predict_params=None, score_params=None, model_id=None,
                 cv_split=None, is_standalone=True):
        """
        Wrapper initialization.

        Parameters
        --------
        :param model: machine learning model
        :param dataset_id: data set identifier
        :param dataset_name: data set name
        :param db_name: database name
        :param fit_params: model's fit parameters
        :param predict_params: model's predict parameters
        :param score_params: model's score parameters
        :param model_id: model identifier
        :param cv_split: cross-validation split identifier
        :param is_standalone: True if model can be used independently from other models
        """
        Wrapper.__init__(self, model, dataset_id, dataset_name, db_name, fit_params,
                         predict_params, score_params, model_id, cv_split, is_standalone)
        self.parameters = self.parse_parameters()

    def load_model(self) -> Any:
        """
        Load model from database.

        Parameters
        --------
        :return: None
        """
        result = self.from_database()
        if result:
            model_id, model_type, model_class, serialized_model, fit_parameters, is_standalone = result
            self.is_fitted = True
            self.model_id = model_id
            self.model = load_neural_model(serialized_model)
            self.serialized_model = serialized_model
        else:
            self.is_fitted = False

    def save_model(self) -> None:
        """
        Save model into database.

        Parameters
        --------
        :return: None
        """
        if not self.serialized_model:
            self.serialized_model = save_neural_model(self.model)
        result = self.to_database()
        if result:
            model_id, model_type, model_class, serialized_model, fit_parameters, is_standalone = result
            self.model_id = model_id

    def set_random_seed(self, seed: int, split_index, random_model):
        """
        Set model random state if possible.

        Parameters
        --------
        :param seed: random seed
        :param split_index: cross-validation split identifier
        :param random_model: whether the model should have the same random state for each cross-validation split
        :return: None
        """
        if random_model:
            random_state = split_index
        else:
            random_state = seed
        tf.set_random_seed(seed=random_state)
        reset_weights(self.model, random_state)
        self.cv_split = split_index
        self.parameters = self.parse_parameters()

    def parse_parameters(self) -> str:
        """
        Parse model parameters.

        Parameters
        --------
        :return: model parameters
        """
        return parse_neural_model(self.model)

    def fit(self, x, y, **kwargs):
        """
        Fit model with some samples.

        Parameters
        --------
        :param x: train data
        :param y: train labels
        :return: None
        """
        if not self.is_fitted:
            fit_params = json.loads(self.fit_parameters)
            self.model.fit(x, y, **fit_params)
            self.is_fitted = True

    def predict(self, x, **kwargs) -> Any:
        """
        Predict labels for some input samples.

        Parameters
        --------
        :param x: input data
        :return: predictions
        """
        predict_params = json.loads(self.predict_parameters)
        return self.model.predict(x, **predict_params)

    def score(self, x, y, **kwargs) -> Any:
        """
        Compute score for some input samples.

        Parameters
        --------
        :param x: input data
        :param y: input labels
        :return: score
        """
        score_params = json.loads(self.score_parameters)
        return self.model.score(x, y, **score_params)


def parse_sklearn_model(model):
    """
    Parse sklearn model parameters.

    Examples
    --------
    >>> import lazygrid as lg
    >>> from sklearn.linear_model import LogisticRegression
    >>>
    >>> lg_model = lg.wrapper.SklearnWrapper(LogisticRegression())
    >>> parameters = lg.wrapper.parse_sklearn_model(lg_model)

    Parameters
    --------
    :param model: sklearn model
    :return: None
    """
    args = []
    for attribute_name in dir(model):
        if not attribute_name.startswith("_") and not attribute_name.endswith("_"):
            try:
                attribute = getattr(model, attribute_name)
                if isinstance(attribute, Callable):
                    attribute = attribute.__name__
                else:
                    attribute = ' '.join(str(attribute).split())
                if attribute_name != str(attribute):
                    args.append(attribute_name + ": " + str(attribute))
            except (sklearn.exceptions.NotFittedError, AttributeError):
                pass
    return ", ".join(args)


def parse_neural_model(model: Model) -> str:
    """
    Parse keras/tensorflow model parameters.

    Examples
    --------
    >>> import lazygrid as lg
    >>>
    >>> keras_nn = lg.neural_models.keras_classifier(layers=[10, 5], input_shape=(20,), n_classes=2)
    >>> parameters = lg.wrapper.parse_neural_model(keras_nn)

    Parameters
    --------
    :param model: neural network model
    :return: None
    """
    parameters = []
    parameters.append("input_shape: " + str(model.input_shape))
    parameters.append("output_shape: " + str(model.output_shape))

    layers = []
    trainable_layers = []
    for layer in model.layers:
        layers.append(layer.output_shape)
        if layer.trainable:
            trainable_layers.append(layer.output_shape)

    parameters.append("layers: " + str(layers))
    parameters.append("trainable_layers: " + str(trainable_layers))

    loss_functions = []
    for item in model.loss_functions:
        loss_functions.append(item.__name__)
    loss_functions.sort()
    parameters.append("loss_functions: " + str(loss_functions))

    losses = []
    for item in model.losses:
        losses.append(item.__name__)
    losses.sort()
    parameters.append("losses: " + str(losses))

    metrics = []
    for item in model.metrics:
        metrics.append(item)
    metrics.sort()
    parameters.append("metrics: " + str(metrics))

    metrics_names = []
    for item in model.metrics_names:
        metrics_names.append(item)
    metrics_names.sort()
    parameters.append("metrics_names: " + str(metrics_names))

    parameters.append("optimizer_class: " + model.optimizer.__class__.__name__)
    optimizer_parameters = []
    for attribute_name in dir(model.optimizer):
        attribute = getattr(model.optimizer, attribute_name)
        if type(attribute) is float:
            optimizer_parameters.append(str(attribute_name) + ": " + str(attribute))
    parameters.append("optimizer_params: " + str(optimizer_parameters))

    return str(parameters)


def corner_cases(model: Any) -> Any:
    """
    Check parameter synonyms.

    Examples
    --------
    >>> import lazygrid as lg
    >>> from sklearn.ensemble import RandomForest
    >>>
    >>> model_1 = lg.wrapper.corner_cases(RandomForestClassifier())
    >>> model_2 = lg.wrapper.corner_cases(RandomForestClassifier(n_estimators=10))
    >>>
    >>> model_1.n_estimators == model_2.n_estimators

    Parameters
    --------
    :param model: sklearn model
    :return: model
    """
    if isinstance(model, RandomForestClassifier):
        if getattr(model, "n_estimators") == "warn":
            setattr(model, "n_estimators", 10)
    return model


def dict_to_json(dictionary: Union[dict, str]) -> str:
    """
    Sort dictionary by key and transform it into a string.

    Examples
    --------
    >>> import lazygrid as lg
    >>>
    >>> my_string = lg.wrapper.dict_to_json({"C": 1, "B": 2, "A": 3})

    Parameters
    --------
    :param dictionary: python dictionary
    :return: sorted dictionary as string
    """
    if isinstance(dictionary, str):
        dictionary = json.loads(dictionary)
    # sort dictionary by key
    if dictionary:
        sorted_dictionary = {}
        for param_key in sorted(dictionary):
            sorted_dictionary[param_key] = dictionary[param_key]
    else:
        sorted_dictionary = {}
    return json.dumps(sorted_dictionary)


def load_neural_model(model_bytes) -> Any:
    """
    Load keras model from binaries.

    Examples
    --------
    >>> import lazygrid as lg
    >>>
    >>> keras_nn = lg.neural_models.keras_classifier(layers=[10, 5], input_shape=(20,), n_classes=2)
    >>>
    >>> tmp_file = "nn_binary.h5"
    >>> keras_nn.save(tmp_file)
    >>> with open(tmp_file, 'rb') as input_file:
    >>>     keras_nn_binary = input_file.read()
    >>>
    >>> keras_nn_recovered = lg.wrapper.load_neural_model(keras_nn_binary)

    Parameters
    --------
    :param model_bytes: serialized keras model
    :return: keras model
    """
    temp = os.path.join("./tmp", "temp.h5")
    with open(temp, 'wb') as output_file:
        output_file.write(model_bytes)
    return keras.models.load_model(temp)


def save_neural_model(model) -> Any:
    """
    Serialize keras model.

    Examples
    --------
    >>> import lazygrid as lg
    >>>
    >>> keras_nn = lg.neural_models.keras_classifier(layers=[10, 5], input_shape=(20,), n_classes=2)
    >>>
    >>> keras_nn_binary = lg.wrapper.save_neural_model(keras_nn)

    Parameters
    --------
    :param model: keras model
    :return: serialized keras model
    """
    if not os.path.isdir("./tmp"):
        os.mkdir("./tmp")
    temp = os.path.join("./tmp", "temp.h5")
    model.save(temp)
    with open(temp, 'rb') as input_file:
        fitted_model = input_file.read()
    return fitted_model
